package edu.berkeley.sbp;
import edu.berkeley.sbp.*;
import edu.berkeley.sbp.Sequence.Position;
import edu.berkeley.sbp.util.*;
import java.io.*;
import java.util.*;
import java.lang.reflect.*;

/** an efficient representation of a collection of trees (Tomita's shared packed parse forest) */
public abstract class Forest<T> /*extends PrintableTree<Forest.Body<T>>*/
    implements IntegerMappable,
               GraphViz.ToGraphViz {

    private static int master_idx = 0;
    private final int idx = master_idx++;
    public int toInt() { return idx; }

    /** assume that this forest contains exactly one tree and return it; otherwise throw an exception */
    public abstract Tree<T> expand1() throws Ambiguous;

    /** expand this forest into a set of trees */
    public void expand(HashSet<Tree<T>> ht) { expand(ht, new HashSet<Forest<T>>(), null); }
    public abstract void expand(HashSet<Tree<T>> ht, HashSet<Forest<T>> ignore, Tree<T> bogus);
    public abstract void gather(HashSet<Forest<T>> ignore);
    public HashSet<Tree<T>> expand(boolean toss) {
        HashSet<Tree<T>> ret = new HashSet<Tree<T>>();
        expand(ret);
        return ret;
    }

    public static <T> Forest<T> create(Input.Region loc, T tag, Forest<T>[] tokens, boolean unwrap) {
        return new Body<T>(loc, tag, tokens, unwrap);
    }
    public abstract void edges(GraphViz.Node n);
    public boolean ambiguous() { return false; }

    // Body //////////////////////////////////////////////////////////////////////////////

    public /*protected*/ static class Body<T> extends Forest<T> /* extends PrintableTree<Forest<T>> implements */ {

        private final Input.Region      location;
        private final T                 tag;
        private final Forest<T>[]       tokens;
        private final boolean           unwrap;

        private Body(Input.Region loc, T tag, Forest<T>[] tokens, boolean unwrap) {
            this.location = loc;
            this.tag = tag;
            this.tokens = tokens==null ? emptyForestArray : new Forest[tokens.length];
            if (tokens != null) System.arraycopy(tokens, 0, this.tokens, 0, tokens.length);
            if (tokens != null) for(int i=0; i<tokens.length; i++) if (tokens[i]==null) throw new Error(i+"");
            this.unwrap = unwrap;
        }

        public Tree<T> expand1() throws Ambiguous {
            Tree<T>[] ret = new Tree[tokens.length];
            for(int i=0; i<tokens.length; i++)
                ret[i] = tokens[i].expand1();
            return new Tree<T>(location, tag, ret, unwrap);
        }
        public void gather(HashSet<Forest<T>> hf) {
            hf.add(this);
            for(Forest<T> f : tokens) f.gather(hf);
        }
        public void expand(HashSet<Tree<T>> ht, HashSet<Forest<T>> ignore, Tree<T> bogus) {
            if (ignore.contains(this)) { ht.add(bogus); return; }
            expand(0, new Tree[tokens.length], ht, ignore, bogus);
        }
        public void expand(final int i, Tree<T>[] ta, HashSet<Tree<T>> ht, HashSet<Forest<T>> ignore, Tree<T> bogus) {
            if (i==tokens.length) {
                ht.add(new Tree<T>(location, tag, ta, unwrap));
            } else {
                HashSet<Tree<T>> ht2 = new HashSet<Tree<T>>();
                tokens[i].expand(ht2, ignore, bogus);
                for(Tree<T> tc : ht2) {
                    ta[i] = tc;
                    expand(i+1, ta, ht, ignore, bogus);
                    ta[i] = null;
                }
            }
        }

        // GraphViz, ToInt //////////////////////////////////////////////////////////////////////////////

        public boolean isTransparent() { return false; }
        public boolean isHidden() { return false; }
        public GraphViz.Node toGraphViz(GraphViz gv) {
            if (gv.hasNode(this)) return gv.createNode(this);
            GraphViz.Node n = gv.createNode(this);
            n.label = headToString()==null?"":headToString();
            n.directed = true;
            edges(n);
            return n;
        }
        boolean edges = false;
        public void edges(GraphViz.Node n) {
            if (edges) return;
            edges = true;
            for(int i=0; i<tokens.length; i++) {
                if (i==tokens.length-1 && unwrap && !tokens[i].ambiguous()) {
                    tokens[i].edges(n);
                } else {
                    n.edge(tokens[i], null);
                }
            }
        }

        protected String  headToString()         { return tag==null?null:tag.toString(); }
        protected String  headToJava()           { return "null"; }
        protected String  left()                 { return "{"; }
        protected String  right()                { return "}"; }
        protected boolean ignoreSingleton()      { return false; }
    }


    // Ref //////////////////////////////////////////////////////////////////////////////

    /**
     *  This class represents a partially complete collection of
     *  forests to be viewed as a forest at some later date; once
     *  viewed, it becomes immutable
     */
    static class Ref<T> extends Forest<T> {
        public HashSet<GSS.Phase.Node> parents = new HashSet<GSS.Phase.Node>();
        private FastSet<Forest<T>> hp = new FastSet<Forest<T>>();

        public Ref() { }

        public Tree<T> expand1() throws Ambiguous {
            if (hp.size() > 1) throw new Ambiguous(this);
            return hp.iterator().next().expand1();
        }
        
        public void gather(HashSet<Forest<T>> ht) {
            ht.add(this);
            for(Forest<T> f : hp) f.gather(ht);
        }

        public Forest resolve() { return this; }
        public void expand(HashSet<Tree<T>> ht, HashSet<Forest<T>> ignore, Tree<T> bogus) {
            if (ignore.contains(this)) { ht.add(bogus); return; }
            for (Forest<T> f : hp) f.expand(ht, ignore, bogus);
        }
        public boolean contains(Forest f) { return hp.contains(f); }
        public void merge(Forest p) { if (p!=this) hp.add(p, true); }
        public boolean ambiguous() {
            if (hp.size()==0) return false;
            if (hp.size()==1) return hp.iterator().next().ambiguous();
            return true;
        }

        // GraphViz, ToInt //////////////////////////////////////////////////////////////////////////////

        public int toInt() {
            if (hp.size()==1) return hp.iterator().next().toInt();
            return super.toInt();
        }
        public boolean isTransparent() { return hp.size()==1; }
        public boolean isHidden() { return hp.size()==0; }
        public void edges(GraphViz.Node n) {
            if (hp.size()==1) { hp.iterator().next().edges(n); return; }
            for(Forest f : hp) f.edges(n);
        }
        public GraphViz.Node toGraphViz(GraphViz gv) {
            //if (hp.size()==0) return null;
            if (hp.size()==1) return hp.iterator().next().toGraphViz(gv);
            if (gv.hasNode(this)) return gv.createNode(this);
            GraphViz.Node n = gv.createNode(this);
            n.label = "?";
            n.color = "red";
            for(Forest f : hp) n.edge(f, null);
            return n;
        }
    }

    // Statics //////////////////////////////////////////////////////////////////////////////

    private static Tree[] tree_hint = new Tree[0];
    private static String[] string_hint = new String[0];
    private static final Forest[] emptyForestArray = new Forest[0];

    protected String  headToString()    { return null; }
    protected String  headToJava()      { return "null"; }
    protected String  left()            { return "<?"; }
    protected String  right()           { return "?>"; }
    protected boolean ignoreSingleton() { return true; }
}
